# -*- coding: utf-8 -*-
'''
Created on Thu September 11 00:07:55 2014

@author: ydiekmann
'''
#
from __future__ import division, print_function
#
import sys
import os
import random
from collections import defaultdict
from string import translate
import numpy as NP
#
from progressbar import ProgressBar, Percentage, Bar, ETA
#
import rpy2.rinterface
rpy2.rinterface.set_initoptions(('rpy2', '--quiet', '--no-save'))
import rpy2.robjects as robjects
R = robjects.r
from rpy2.robjects.packages import importr
grdevices = importr('grDevices')
R.library('LaplacesDemon')
#
#
#
BASE = '/Users/ydiekmann/Documents/__MACE/6__Gibbs_age_method/'
DATA = BASE + 'data/'
TMP = BASE + 'tmp/'
#
#
#
widgets = ['      ', Percentage(), ' ', Bar(marker='.'), ' ', ETA()] # Bar(marker=RotatingMarker())
#
#
#
# *****************************
# *          CLASSES          *
# *****************************
#
# http://stackoverflow.com/questions/6796492/python-temporarily-redirect-stdout-stderr
class Turn_silent(object):
    #
    DEVNULL = open(os.devnull, 'w')
    #
    def __enter__(self):
        sys.stdout.flush()
        sys.stderr.flush()
        self.old_stdout, self.old_stderr, sys.stdout, sys.stderr = \
                                                        sys.stdout, sys.stderr, Turn_silent.DEVNULL, Turn_silent.DEVNULL
    #
    def __exit__(self, exc_type, exc_value, traceback):
        Turn_silent.DEVNULL.flush()
        sys.stdout, sys.stderr = self.old_stdout, self.old_stderr
#
#
#
class Rank_node(object):
    """
    Class defining a node in a rank graph.
    
    :arg config: 
    :type config: string
    """
    #
    def __init__(self, idi, min_age, max_age, overwrite=False):#, config=):
        self.idi = idi
        self.min_age = min_age
        self.max_age = max_age
        self.burnin = None
        #
        self.child_node = None
        #
        self.current_value = None
        self.age_samples = []
    #
    #
    #
    def __flush_age_samples(self):
        with open(self.age_samples_file_name, 'a') as to_write:
            #
            if os.path.getsize(self.age_samples_file_name) > 0:
                to_write.write(',')
            #
            to_write.write('%s,' % translate(str(self.age_samples[:-1]), None, '[] '))
            to_write.write('%f' % self.age_samples[-1])
        #
        self.age_samples = []
    #
    def add_age_sample(self, age_sample):
        self.age_samples.append(age_sample)
        #
        if len(self.age_samples) == 1000:
            self.__flush_age_samples()
    #
    #
    #
    def initialise(self, burnin, parent_value, child, rank_id):
        #
        self.burnin = burnin
        #
        self.age_samples = []
        self.current_value = max(parent_value, self.min_age)
        #
        self.child_node = child
        #
        self.age_samples_file_name = TMP + '%s_rank%s.csv' % (self.idi, rank_id)
        #
        if os.path.exists(self.age_samples_file_name):
            open(self.age_samples_file_name, 'w').close()
        #
        return self.current_value
    #
    #
    #
    def gibbs_scan(self, parent_value=0):
        #
        lower_bound = max(self.min_age, parent_value)
        #
        if self.child_node is not None:
            upper_bound = min(self.max_age, self.child_node.current_value) #self.child_node.age_samples[-1])
        else:
            upper_bound = self.max_age
        #
        assert lower_bound <= upper_bound
        self.current_value = random.uniform(lower_bound, upper_bound)
        #
        if self.burnin:
            self.burnin -= 1
        else:
            self.add_age_sample(self.current_value)#, repetitions)
        #
        if self.child_node is not None:
            self.child_node.gibbs_scan(parent_value=self.current_value)
#
#
#
# *****************************
# *         FUNCTIONS         *
# *****************************
#
# only for data in the following format...
'''
ID;Lower Bound;Upper Bound
1;25,00;35,00
2;25,00;38,00
3;6,00;12,00
'''
def parse_rank_nodes(rank_node_file_name, overwrite=False):
    idi_to_rank_node = {}
    #
    with open(rank_node_file_name, 'rU') as to_parse:
        to_parse.next()
        #
        for line in to_parse:
            idi, min_age, max_age = line.strip().split(';')
            #
            idi_to_rank_node[idi] = Rank_node(idi, float(min_age.replace(',', '.')), \
                                                    float(max_age.replace(',', '.')), overwrite=overwrite)
    #
    return idi_to_rank_node
#
#
#
# only for data in the following format... from SMALLEST to BIGGEST ?!
'''
1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22;23;24;25;26;27;28;29;30;31;32;33;34;35;36;37;38;39;40;41;42;43
34;11;10;24;1;49;16;15;13;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
5;4;3;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
'''
def __parse_rankings(rankings_file_name, pos_list):
    filtered_individuals = set()
    list_of_rankings = []
    with open(rankings_file_name, 'rU') as to_parse:
        for line in to_parse:
            ranking = [el for el in line.strip().split(';') if el and ('/' in el or el in pos_list)]
            # just to see which ones are filtered out!
            filtered_individuals.update([el for el in line.strip().split(';') if el and \
                                                                                not '/' in el and not el in pos_list])
            if len(ranking) > 1:
                list_of_rankings.append(ranking)
    return list_of_rankings, filtered_individuals
#
def parse_birth_orders(birth_orders_file_name, pos_list):
    return __parse_rankings(birth_orders_file_name, pos_list)
#
def parse_age_ranks(age_ranks_file_name, pos_list):
    return __parse_rankings(age_ranks_file_name, pos_list)
#
#
#
def age_rank_conflicts_birth_orders(age_rank, list_of_birth_orders):
    #
    conflicting_idis = set()
    idi_pairs_to_birth_orders = defaultdict(list)
    for birth_order in list_of_birth_orders:
        #
        indices = [age_rank.index(idi) for idi in birth_order if idi in age_rank]
        #
        for i in xrange(1, len(indices)):
            if indices[i-1] > indices[i]:
                #return birth_order
                conflicting_idis.add(age_rank[indices[i-1]])
                conflicting_idis.add(age_rank[indices[i]])
                idi_pairs_to_birth_orders[(age_rank[indices[i-1]], age_rank[indices[i]])].append(birth_order)
    #
    return conflicting_idis, idi_pairs_to_birth_orders
#
#
#
def age_rank_conflicts_interval_constraints(age_rank, idi_to_rank_node): #, conflict_pairs=None
    #
    idi_to_nb_of_conflicts = defaultdict(int)
    for i in xrange(len(age_rank)-1):
        for j in xrange(i+1, len(age_rank)):
            if idi_to_rank_node[age_rank[i]].min_age > idi_to_rank_node[age_rank[j]].max_age:
                idi_to_nb_of_conflicts[age_rank[i]] += 1
                idi_to_nb_of_conflicts[age_rank[j]] += 1
    #
    max_idi, max_conflicts = '', 0
    for idi, nb_of_conflicts in idi_to_nb_of_conflicts.iteritems():
        if nb_of_conflicts > max_conflicts:                             # TODO this means that always the first one
            max_conflicts = nb_of_conflicts                             # is removed; make it random
            max_idi = idi
    #
    if max_conflicts > 0:
        age_rank.remove(max_idi)
        age_rank = age_rank_conflicts_interval_constraints(age_rank, idi_to_rank_node)
    #
    return age_rank
#
#
#
def redundant_age_rank(age_rank_1, age_rank_2):
    # as equal ones have been removed before, redundant ones must have unequal size
    if len(age_rank_1) == len(age_rank_2):
        return None
    # 
    if len(age_rank_1) > len(age_rank_2):
        age_rank_long, age_rank_short = age_rank_1, age_rank_2
    else:
        age_rank_long, age_rank_short = age_rank_2, age_rank_1
    #
    short_indices = [age_rank_long.index(idi) if idi in age_rank_long else None for idi in age_rank_short]
    #
    if None in short_indices:
        return None
    #
    for i in xrange(len(short_indices)-1):
        if short_indices[i] > short_indices[i+1]:
            return None
    #
    return (age_rank_short, age_rank_long)
#
#
#
def merge_age_ranks(age_rank_1, indices_1, age_rank_2, indices_2):
    assert (indices_1[0] is None and indices_2[0] is not None) or indices_1[0] <= indices_2[0]
    #
    age_rank = []
    #
    i1, i2 = 0, 0
    for k in xrange(len(age_rank_1) + len(age_rank_2)):
        #
        if i2==len(indices_2)-1:
            age_rank.append(age_rank_2[i2])
            break
        #
        if i1 < len(age_rank_1) and indices_1[i1] is not None and i2 < len(age_rank_2) and indices_2[i2] is not None:
            assert age_rank_2[indices_1[i1]] == age_rank_1[indices_2[i2]]
            #
            age_rank.append(age_rank_2[indices_1[i1]])
            #
            i1 += 1
            i2 += 1
            #
            continue
        #
        # ORDER IS IMPORTANT! As I know that age_rank_1's end is going to be reached earlier or at the same time
        if i1==len(age_rank_1) or (indices_1[i1] is not None and indices_2[i2] is None):
            #
            age_rank.append(age_rank_2[i2])
            #
            i2 += 1
            #
            continue
        #
        if indices_1[i1] is None and indices_2[i2] is not None:
            #
            age_rank.append(age_rank_1[i1])
            #
            i1 += 1
            #
            continue
    #
    return age_rank
#
#
#
# *****************************
# *           MAIN            *
# *****************************
#
if __name__ == "__main__":
    #
    print('\n\n   ***  GIBBS SAMPLING AGE RANKS  ***')
    logfile = open(TMP + 'log.txt', 'w')
    logfile.write('\n\n   ***  GIBBS SAMPLING AGE RANKS  ***\n')
    #
    #
    #
    print('\n   Parse input data')
    logfile.write('\n\n   Parse input data\n')
    #
    print('   -- age brackets...', end='')
    idi_to_rank_node = parse_rank_nodes(DATA + 'Age_brackets.csv', overwrite=False)
    print(' (#%d) DONE' % len(idi_to_rank_node))
    #
    print('   -- birth orders...', end='')
    list_of_birth_orders, filtered_individuals = parse_birth_orders(\
                                                            DATA + 'Birth_orders.csv', idi_to_rank_node.keys())
    print(' (#%d) DONE' % len(list_of_birth_orders))
    logfile.write('\n      filtered individuals from birth orders that have no age bracket:\n      %s\n' % \
                                                                                        str(list(filtered_individuals)))
    #
    print('   -- age ranks...', end='')
    logfile.write('\n   -- age ranks\n')
    list_of_age_ranks, filtered_individuals = parse_age_ranks(DATA + 'Ranking_orders.csv', idi_to_rank_node.keys())
    print(' (#%d) DONE' % len(list_of_age_ranks))
    logfile.write('\n      filtered individuals from age ranks that have no age bracket:\n      %s\n' % \
                                                                                        str(list(filtered_individuals)))
    #
    print('   -- merge age ranks and birth orders...', end='')
    list_of_age_ranks.extend(list_of_birth_orders)
    print(' (#%d) DONE' % len(list_of_age_ranks))
    #
    #
    #
    print('\n   Filter ranks and rank nodes')
    logfile.write('\n   Filter ranks and rank nodes\n')
    #
    print('   -- filter age ranks by birth orders...', end='')
    logfile.write('\n   -- filter age ranks by birth orders\n')
    _filtered_list_of_age_ranks = []
    for age_rank in list_of_age_ranks:
        conflicting_idis, idi_pairs_to_birth_orders = age_rank_conflicts_birth_orders(age_rank, list_of_birth_orders)
        #
        if not conflicting_idis:
            pass
        else:
            logfile.write('\n      age rank\n      %s\n      has conflicts' % str(age_rank))
            for conflicting_idi in conflicting_idis:
                age_rank.remove(conflicting_idi)
            for idi_pairs, birth_orders in idi_pairs_to_birth_orders.iteritems():
                logfile.write('\n        %s: %s' % (str(idi_pairs), str(birth_orders)))
            logfile.write('\n      new age rank\n      %s\n' % str(age_rank))
        _filtered_list_of_age_ranks.append(age_rank)
    print('(#%d) DONE' % len(_filtered_list_of_age_ranks))
    #
    print('   -- filter age ranks by age brackets...', end='')
    logfile.write('\n   -- filter age ranks by age brackets\n')
    filtered_list_of_age_ranks = []
    for age_rank in _filtered_list_of_age_ranks:
        filtered_age_rank = age_rank_conflicts_interval_constraints(list(age_rank), idi_to_rank_node)
        filtered_list_of_age_ranks.append(filtered_age_rank)
        if filtered_age_rank != age_rank:
            logfile.write('\n      age rank\n      %s\n      conflicts age brackets; new rank\n      %s\n' % \
                                                                                (str(age_rank), str(filtered_age_rank)))
    print(' (#%d) DONE' % len(filtered_list_of_age_ranks))
    #
    print('   -- filter duplicate age ranks...', end='')
    logfile.write('\n   -- filter duplicate age ranks\n')
    age_ranks_to_delete = []
    for i in xrange(len(filtered_list_of_age_ranks)-1):
        for j in xrange(i+1, len(filtered_list_of_age_ranks)):
            if filtered_list_of_age_ranks[i]==filtered_list_of_age_ranks[j]:
                age_ranks_to_delete.append(filtered_list_of_age_ranks[i])
                break
    for age_rank in age_ranks_to_delete:
        filtered_list_of_age_ranks.remove(age_rank)
        logfile.write('\n      age rank is a duplicate\n      %s\n' % str(age_rank))
    print(' (#%d) DONE' % len(filtered_list_of_age_ranks))
    #
    # filters also those age_ranks that are a subsequence of some other, otherwise the merge rank function breaks
    print('   -- filter redundant age ranks...', end='')
    logfile.write('\n   -- filter redundant age ranks\n')
    age_rank_pairs_to_delete = []
    for i in xrange(len(filtered_list_of_age_ranks)-1):
        for j in xrange(i+1, len(filtered_list_of_age_ranks)):
            age_rank_pair = redundant_age_rank(filtered_list_of_age_ranks[i], filtered_list_of_age_ranks[j])
            if age_rank_pair is None:
                continue
            else:
                age_rank_pairs_to_delete.append(age_rank_pair)
    for age_rank_short, age_rank_long in age_rank_pairs_to_delete:
        try:
            filtered_list_of_age_ranks.remove(age_rank_short)
            logfile.write('\n      age rank\n      %s\n      is redundant for example to\n      %s\n' % \
                                                                            (str(age_rank_short), str(age_rank_long)))
        except ValueError:
            pass
    print(' (#%d) DONE' % len(filtered_list_of_age_ranks))
    #
    print('   -- establish remaining rank nodes...', end='')
    filtered_idis = set()
    for age_rank in filtered_list_of_age_ranks:
        for idi in age_rank:
            filtered_idis.add(idi)
    print(' (#%d) DONE' % len(filtered_idis))
    #
    print('   -- remove filtered rank nodes...', end='')
    filtered_rank_nodes = set()
    for idi in idi_to_rank_node.keys():     # .keys() is needed as dic moified during traversal
        if idi not in filtered_idis:
            filtered_rank_nodes.add(idi)
            del idi_to_rank_node[idi]
    print(' (#%d) DONE' % len(idi_to_rank_node))
    #
    print('   -- write filtered rank nodes...', end='')
    logfile.write('\n   -- write filtered individuals:\n      ')
    filtered_rank_nodes = list(filtered_rank_nodes)
    if filtered_rank_nodes:
        for idi in filtered_rank_nodes[:-1]:
            logfile.write('%s, ' % idi)
        logfile.write('%s\n' % filtered_rank_nodes[-1])
    print(' DONE')
    #
    #
    #
    print('\n   Merge age ranks')
    logfile.write('\n   -- Merged age ranks\n')
    #
    counter = 0
    while len(filtered_list_of_age_ranks)>1:
        print('   -- round %d (# age ranks: %d)' % ((counter + 1), len(filtered_list_of_age_ranks)))
        #
        involved_indices = set()
        index_pairs = []
        #
        _pairs = (len(filtered_list_of_age_ranks) * (len(filtered_list_of_age_ranks)-1)) / 2
        print('      pairwise testing (%d comparisons)' % _pairs)
        pbar = ProgressBar(widgets=widgets, maxval=_pairs).start()
        
        _counter = 0
        for i in xrange(len(filtered_list_of_age_ranks)-1):
            for j in xrange(i+1, len(filtered_list_of_age_ranks)):
                _counter += 1
                pbar.update(_counter)
                #
                i_age_rank = filtered_list_of_age_ranks[i]
                j_age_rank = filtered_list_of_age_ranks[j]
                #
                i_indices = [j_age_rank.index(idi) if idi in j_age_rank else None for idi in i_age_rank]
                j_indices = [i_age_rank.index(idi) if idi in i_age_rank else None for idi in j_age_rank]
                #
                _i_indices = [k for k in i_indices if k is not None]
                _j_indices = [k for k in j_indices if k is not None]
                #
                # if any of these are true no merging is possible
                try:
                    # TODO is there a case of only Nones?!
                    #
                    assert all([_i_indices[k-1] < _i_indices[k] for k in xrange(1, len(_i_indices))])
                    assert all([_j_indices[k-1] < _j_indices[k] for k in xrange(1, len(_j_indices))])
                    #
                    assert not (i_indices[0] is None and j_indices[0] is None)
                    assert not (i_indices[-1] is None and j_indices[-1] is None)
                #
                except AssertionError:
                    continue
                #
                __i_indices = [i_indices.index(k) for k in i_indices if k is not None]
                __j_indices = [j_indices.index(k) for k in j_indices if k is not None]
                #
                try:
                    for k in xrange(len(__i_indices)-1):
                        if __i_indices[k+1] - __i_indices[k] > 1:
                            assert i_indices[__i_indices[k+1]] - i_indices[__i_indices[k]] == 1
                    for k in xrange(len(__j_indices)-1):
                        if __j_indices[k+1] - __j_indices[k] > 1:
                            assert j_indices[__j_indices[k+1]] - j_indices[__j_indices[k]] == 1
                except AssertionError:
                    continue
                #
                #
                if (i_indices[0] is None and j_indices[0] is not None) or \
                            ((i_indices[0] is not None and j_indices[0] is not None) and i_indices[0] <= j_indices[0]):
                    _tmp_merged = merge_age_ranks(i_age_rank, i_indices, j_age_rank, j_indices)
                else:
                    _tmp_merged = merge_age_ranks(j_age_rank, j_indices, i_age_rank, i_indices)
                #
                try:
                    for _i in xrange(len(_tmp_merged)-1):
                        for _j in xrange(_i+1, len(_tmp_merged)):
                            if idi_to_rank_node[_tmp_merged[_i]].min_age > idi_to_rank_node[_tmp_merged[_j]].max_age:
                                raise AssertionError
                except AssertionError:
                    continue
                #
                #
                involved_indices.add(i)
                involved_indices.add(j)
                index_pairs.append((i, j, i_indices, j_indices))
                #
                # THIS BREAK GENERATES ONLY 1 RANDOM PAIRING AND NOT ALL POSSIBLE ONES !!
                # corresponds to greedy strategy! slightly more invovled than "not merging" strategy described in paper
                break
        #
        pbar.finish()
        #
        #
        if not index_pairs:
            break
        #
        #
        print('      merging (%d pairs)' % len(index_pairs))
        pbar = ProgressBar(widgets=widgets, maxval=len(index_pairs)).start()
        #
        merged_filtered_list_of_age_ranks = []
        for _pbar, (i, j, i_indices, j_indices) in enumerate(index_pairs):
            pbar.update(_pbar)
            #
            i_age_rank = filtered_list_of_age_ranks[i]
            j_age_rank = filtered_list_of_age_ranks[j]
            #
            if (i_indices[0] is None and j_indices[0] is not None) or \
                            ((i_indices[0] is not None and j_indices[0] is not None) and i_indices[0] <= j_indices[0]):
                merged_filtered_list_of_age_ranks.append(merge_age_ranks(i_age_rank, i_indices, j_age_rank, j_indices))
            else:
                merged_filtered_list_of_age_ranks.append(merge_age_ranks(j_age_rank, j_indices, i_age_rank, i_indices))
        pbar.finish()
        #
        #
        for index, age_rank in enumerate(filtered_list_of_age_ranks):
            if index not in involved_indices:
                merged_filtered_list_of_age_ranks.append(age_rank)
        #
        filtered_list_of_age_ranks = merged_filtered_list_of_age_ranks
        #
        counter += 1
    #
    #
    for age_rank in filtered_list_of_age_ranks:
        logfile.write('\n      %s\n' % str(age_rank))
    #
    #
    #
    print('\n   Run Gibbs sampler')
    #
    for index, age_rank in enumerate(filtered_list_of_age_ranks):
        #
        BURNIN = 50
        print('   -- ' + ('%d' % (index+1)).rjust(2) + '/%d initialise' % len(filtered_list_of_age_ranks), end='')
        parent_value = 0
        for i in xrange(len(age_rank)-1):
            parent_value = idi_to_rank_node[age_rank[i]].initialise(\
                                                    BURNIN, parent_value, idi_to_rank_node[age_rank[i+1]], str(index))
        idi_to_rank_node[age_rank[-1]].initialise(BURNIN, parent_value, None, str(index))
        #
        REPS = 15000 + BURNIN
        print(', scan:\n', end='')
        pbar = ProgressBar(widgets=widgets, maxval=REPS).start()
        for i in xrange(REPS):
            pbar.update(i)
            idi_to_rank_node[age_rank[0]].gibbs_scan()
        pbar.finish()
    #
    #
    #
    
    
    
    
    
    # TODO
    # merge samples belonging to one individual across different rank orders
    #
    
    logfile.close()
    print('\n\n   END\n')
    sys.exit()
    
    
    
    #
    #
    #
    print('\n   Plot distributions')
    logfile.write('\n   -- Final distributions\n')
    #
    for idi, rank_node in idi_to_rank_node.iteritems():
        if not os.path.exists(TMP + '%s.csv' % idi) or os.path.getsize(TMP + '%s.csv' % idi) == 0:
            continue
        #
        print('   -- %s' % idi)
        logfile.write('\n      individual %s:' % idi)
        #
        if rank_node.min_age==rank_node.max_age:
            print('      # no range: %.2f' % rank_node.min_age)
            logfile.write('\n      # no range:  %.2f\n' % rank_node.min_age)
        else:
            #
            with Turn_silent():
                R("data = scan(file='%s%s.csv', sep=',', what = double())" % (TMP, idi))
            data = R("data")
            #
            R("dens = density(data)")
            R("ymax = max(dens$y)")
            #
            grdevices.pdf(file=TMP + '%s.pdf' % idi)
            if min(data) != max(data):
                R("plot(dens, xlim=c(%f, %f), ylim=c(-0.1*ymax, ymax), main='Individual %s')" % \
                                                                            (rank_node.min_age, rank_node.max_age, idi))
                R("lines(c(%f, %f), c(-0.05*ymax, -0.05*ymax))" % (rank_node.min_age, rank_node.max_age))
                #
                with Turn_silent():
                    # http://artax.karlin.mff.cuni.cz/r-help/library/LaplacesDemon/html/p.interval.html
                    R("hpd = p.interval(data, HPD=TRUE, MM=TRUE, prob=0.95, plot=FALSE, PDF=FALSE)")
                hpd_low, hpd_high = R("as.numeric(hpd[1, ])")
            #
            else:
                R("plot(c(data[1], data[1]), c(0, 1), xlim=c(%f, %f), ylim=c(-0.1, 1), type='l', "
                  "xlab='', ylab='Density', main='Individual %s')" % (rank_node.min_age, rank_node.max_age, idi))
                R("lines(c(%f, %f), c(-0.05, -0.05))" % (rank_node.min_age, rank_node.max_age))
                #
                hpd_low, hpd_high = data[0], data[0]
            grdevices.dev_off()
            #
            mode = R("dens$x[which.max(dens$y)]")[0]
            #
            print('      # range:  %.2f-%.2f' % (rank_node.min_age, rank_node.max_age))
            print('      # mean:   %f' % NP.mean(data))
            print('      # median: %f' % NP.median(data))
            print('      # mode:   %f' % mode)
            print('      # 95%%HPD: %.2f-%.2f' % (hpd_low, hpd_high))
            #
            logfile.write('\n      # range:  %.2f-%.2f\n      # mean:   %f\n      # median: %f\n      # mode:   %f\n'
                                '      # 95%%HPD: %.2f-%.2f\n' % (rank_node.min_age, rank_node.max_age, NP.mean(data), \
                                                                  NP.median(data), mode, hpd_low, hpd_high))
    #
    #
    #
    logfile.close()
    print('\n\n   END\n')
    sys.exit()
#
#
#


